import asyncio
import datetime as dt
import logging
from collections import Counter
from contextlib import contextmanager
from functools import partial, wraps
from typing import Mapping, Optional, Sequence

import qtawesome as qta
import yaml
from ophyd import Device as ThreadedDevice
from ophyd_async.core import Device
from pydm import PyDMChannel
from qasync import asyncSlot
from qtpy.QtCore import QDateTime, Qt
from qtpy.QtGui import QStandardItem, QStandardItemModel
from tiled.client.container import Container

from firefly import display
from firefly.run_browser.client import DatabaseWorker
from firefly.run_browser.widgets import ExportDialog

log = logging.getLogger(__name__)


def cancellable(fn):
    @wraps(fn)
    async def inner(*args, **kwargs):
        try:
            return await fn(*args, **kwargs)
        except asyncio.exceptions.CancelledError:
            log.warning(f"Cancelled task {fn}")

    return inner


class RunBrowserDisplay(display.FireflyDisplay):
    runs_model: QStandardItemModel
    _run_col_names: Sequence = [
        "Plan",
        "Sample",
        "Edge",
        "E0",
        "Exit Status",
        "Datetime",
        "UID",
        "Proposal",
        "ESAF",
        "ESAF Users",
    ]
    _multiplot_items = {}

    selected_runs: list
    _running_db_tasks: Mapping

    proposal_channel: PyDMChannel
    esaf_channel: PyDMChannel

    export_dialog: Optional[ExportDialog] = None

    # Counter for keeping track of UI hints for long DB hits
    _busy_hinters: Counter

    def __init__(self, args=None, macros=None, **kwargs):
        super().__init__(args=args, macros=macros, **kwargs)
        self.selected_runs = []
        self._running_db_tasks = {}
        self._busy_hinters = Counter()
        self.reset_default_filters()

    async def setup_database(self, tiled_client: Container, catalog_name: str):
        """Prepare to use a set of databases accessible through *tiled_client*.

        Parameters
        ==========
        Each key in *tiled_client* should be"""
        self.db = DatabaseWorker(tiled_client)
        self.ui.catalog_combobox.addItems(await self.db.catalog_names())
        self.ui.catalog_combobox.setCurrentText(catalog_name)
        await self.change_catalog(catalog_name)

    @asyncSlot(str)
    async def change_catalog(self, catalog_name: str):
        """Activate a different catalog in the Tiled server."""
        await self.db_task(self.db.change_catalog(catalog_name), name="change_catalog")
        await self.db_task(
            asyncio.gather(self.load_runs(), self.update_combobox_items()),
            name="change_catalog",
        )

    def db_task(self, coro, name="default task"):
        """Executes a co-routine as a database task. Existing database
        tasks get cancelled.

        """
        # Check for existing tasks
        has_previous_task = name in self._running_db_tasks.keys()
        task_is_running = has_previous_task and not self._running_db_tasks[name].done()
        if task_is_running:
            self._running_db_tasks[name].cancel("New database task started.")
        # Wait on this task to be done
        new_task = asyncio.ensure_future(coro)
        self._running_db_tasks[name] = new_task
        return new_task

    @asyncSlot()
    async def reload_runs(self):
        """A simple wrapper to make load_runs a slot."""
        await self.load_runs()

    @cancellable
    async def load_runs(self):
        """Get the list of available runs based on filters."""
        with self.busy_hints(run_widgets=True, run_table=True, filter_widgets=False):
            runs = await self.db_task(
                self.db.load_all_runs(self.filters()),
                name="load all runs",
            )
            # Update the table view data model
            self.runs_model.clear()
            self.runs_model.setHorizontalHeaderLabels(self._run_col_names)
            for run in runs:
                items = [QStandardItem(val) for val in run.values()]
                self.ui.runs_model.appendRow(items)
            # Adjust the layout of the data table
            sort_col = self._run_col_names.index("Datetime")
            self.ui.run_tableview.sortByColumn(sort_col, Qt.DescendingOrder)
            self.ui.run_tableview.resizeColumnsToContents()
            # Let slots know that the model data have changed
            self.runs_total_label.setText(str(self.ui.runs_model.rowCount()))

    def clear_filters(self):
        self.ui.filter_plan_combobox.setCurrentText("")
        self.ui.filter_sample_combobox.setCurrentText("")
        self.ui.filter_formula_combobox.setCurrentText("")
        self.ui.filter_edge_combobox.setCurrentText("")
        self.ui.filter_exit_status_combobox.setCurrentText("")
        self.ui.filter_user_combobox.setCurrentText("")
        self.ui.filter_proposal_combobox.setCurrentText("")
        self.ui.filter_esaf_combobox.setCurrentText("")
        self.ui.filter_current_proposal_checkbox.setChecked(False)
        self.ui.filter_current_esaf_checkbox.setChecked(False)
        self.ui.filter_beamline_combobox.setCurrentText("")
        self.ui.filter_after_checkbox.setChecked(False)
        self.ui.filter_before_checkbox.setChecked(False)
        self.ui.filter_full_text_lineedit.setText("")
        self.ui.filter_standards_checkbox.setChecked(False)

    def reset_default_filters(self):
        self.clear_filters()
        self.ui.filter_exit_status_combobox.setCurrentText("success")
        self.ui.filter_current_esaf_checkbox.setChecked(True)
        self.ui.filter_current_proposal_checkbox.setChecked(True)
        self.ui.filter_after_checkbox.setChecked(True)
        last_week = dt.datetime.now().astimezone() - dt.timedelta(days=7)
        last_week = QDateTime.fromTime_t(int(last_week.timestamp()))
        self.ui.filter_after_datetimeedit.setDateTime(last_week)

    async def update_combobox_items(self):
        """"""
        with self.busy_hints(run_table=False, run_widgets=False, filter_widgets=True):
            fields = await self.db.load_distinct_fields()
            for field_name, cb in [
                ("plan_name", self.ui.filter_plan_combobox),
                ("sample_name", self.ui.filter_sample_combobox),
                ("sample_formula", self.ui.filter_formula_combobox),
                ("edge", self.ui.filter_edge_combobox),
                ("exit_status", self.ui.filter_exit_status_combobox),
                ("proposal_id", self.ui.filter_proposal_combobox),
                ("esaf_id", self.ui.filter_esaf_combobox),
                ("beamline_id", self.ui.filter_beamline_combobox),
            ]:
                if field_name in fields.keys():
                    old_text = cb.currentText()
                    cb.clear()
                    cb.addItems(fields[field_name])
                    cb.setCurrentText(old_text)

    def customize_ui(self):
        self.load_models()
        # Setup controls for select which run to show
        self.ui.run_tableview.selectionModel().selectionChanged.connect(
            self.update_selected_runs
        )
        self.ui.refresh_runs_button.setIcon(qta.icon("fa5s.sync"))
        self.ui.refresh_runs_button.clicked.connect(self.reload_runs)
        self.ui.reset_filters_button.clicked.connect(self.reset_default_filters)
        # Respond to changes in displaying the 1d plot
        for signal in [
            self.ui.signal_y_combobox.currentTextChanged,
            self.ui.signal_x_combobox.currentTextChanged,
            self.ui.signal_r_combobox.currentTextChanged,
            self.ui.signal_r_checkbox.stateChanged,
            self.ui.logarithm_checkbox.stateChanged,
            self.ui.invert_checkbox.stateChanged,
            self.ui.gradient_checkbox.stateChanged,
        ]:
            signal.connect(self.plot_1d_view.clear_runs)
            signal.connect(self.update_1d_plot)
        self.ui.plot_1d_hints_checkbox.stateChanged.connect(self.update_1d_signals)
        self.ui.autorange_1d_button.clicked.connect(self.auto_range)
        # Respond to changes in displaying the 2d plot
        for signal in [
            self.ui.plot_multi_hints_checkbox.stateChanged,
            self.ui.multi_signal_x_combobox.currentTextChanged,
        ]:
            signal.connect(self.update_multi_signals)
            signal.connect(self.update_multi_plot)
        # Respond to changes in displaying the 2d plot
        self.ui.signal_value_combobox.currentTextChanged.connect(self.update_2d_plot)
        self.ui.logarithm_checkbox_2d.stateChanged.connect(self.update_2d_plot)
        self.ui.invert_checkbox_2d.stateChanged.connect(self.update_2d_plot)
        self.ui.gradient_checkbox_2d.stateChanged.connect(self.update_2d_plot)
        self.ui.plot_2d_hints_checkbox.stateChanged.connect(self.update_2d_signals)
        # Select a new catalog
        self.ui.catalog_combobox.currentTextChanged.connect(self.change_catalog)
        # Respond to filter controls getting updated
        self.ui.filters_widget.returnPressed.connect(self.refresh_runs_button.click)
        # Respond to controls for the current run
        self.ui.export_button.clicked.connect(self.export_runs)
        self.ui.reload_plots_button.clicked.connect(self.update_plots)
        # Set up 1D plotting widgets
        self.plot_1d_item = self.ui.plot_1d_view.getPlotItem()
        self.plot_2d_item = self.ui.plot_2d_view.getImageItem()
        self.plot_1d_item.addLegend()
        self.plot_1d_item.hover_coords_changed.connect(
            self.ui.hover_coords_label.setText
        )
        # Create a new export dialog for saving files
        self.export_dialog = ExportDialog(parent=self)

    async def update_devices(self, registry):
        try:
            bss_device = registry["bss"]
        except KeyError:
            log.warning("Could not find device 'bss', disabling 'current' filters.")
            self.ui.filter_current_proposal_checkbox.setChecked(False),
            self.ui.filter_current_proposal_checkbox.setEnabled(False),
            self.ui.filter_current_esaf_checkbox.setChecked(False),
            self.ui.filter_current_esaf_checkbox.setEnabled(False),
        else:
            self.setup_bss_channels(bss_device)
        await super().update_devices(registry)

    def setup_bss_channels(self, bss: Device | ThreadedDevice):
        """Setup channels to update the proposal and ESAF ID boxes."""
        if getattr(self, "proposal_channel", None) is not None:
            self.proposal_channel.disconnect()
        self.proposal_channel = PyDMChannel(
            address=f"haven://{bss.proposal.proposal_id.name}",
            value_slot=partial(
                self.update_bss_filter,
                combobox=self.ui.filter_proposal_combobox,
                checkbox=self.ui.filter_current_proposal_checkbox,
            ),
        )
        if getattr(self, "esaf_channel", None) is not None:
            self.esaf_channel.disconnect()
        self.esaf_channel = PyDMChannel(
            address=f"haven://{bss.esaf.esaf_id.name}",
            value_slot=partial(
                self.update_bss_filter,
                combobox=self.ui.filter_esaf_combobox,
                checkbox=self.ui.filter_current_esaf_checkbox,
            ),
        )

    def update_bss_filter(self, text: str, *, combobox, checkbox):
        """If *checkbox* is checked, update *combobox* with new *text*."""
        if checkbox.checkState():
            combobox.setCurrentText(text)

    def auto_range(self):
        self.plot_1d_view.autoRange()

    def update_busy_hints(self):
        """Enable/disable UI elements based on the active hinters."""
        # Widgets for showing plots for runs
        if self._busy_hinters["run_widgets"] > 0:
            self.ui.detail_tabwidget.setEnabled(False)
        else:
            # Re-enable the run widgets
            self.ui.detail_tabwidget.setEnabled(True)
        # Widgets for selecting which runs to show
        if self._busy_hinters["run_table"] > 0:
            self.ui.run_tableview.setEnabled(False)
        else:
            # Re-enable the run widgets
            self.ui.run_tableview.setEnabled(True)
        # Widgets for filtering runs
        if self._busy_hinters["filters_widget"] > 0:
            self.ui.filters_widget.setEnabled(False)
        else:
            self.ui.filters_widget.setEnabled(True)
        # Update status message in message bars
        if len(list(self._busy_hinters.elements())) > 0:
            self.show_message("Loading…")
        else:
            self.show_message("Done.", 5000)

    @contextmanager
    def busy_hints(self, run_widgets=True, run_table=True, filter_widgets=True):
        """A context manager that displays UI hints when slow operations happen.

        Arguments can be used to control which widgets are modified.

        Usage:

        .. code-block:: python

            with self.busy_hints():
                self.db_task(self.slow_operation)

        Parameters
        ==========
        run_widgets
          Disable the widgets for viewing individual runs.
        run_table
          Disable the table for selecting runs to view.
        filter_widgets
          Disable the filter comboboxes, etc.

        """
        # Update the counters for keeping track of concurrent contexts
        hinters = {
            "run_widgets": run_widgets,
            "run_table": run_table,
            "filters_widget": filter_widgets,
        }
        hinters = [name for name, include in hinters.items() if include]
        self._busy_hinters.update(hinters)
        # Update the UI (e.g. disable widgets)
        self.update_busy_hints()
        # Run the innner context code
        try:
            yield
        finally:
            # Re-enable widgets if appropriate
            self._busy_hinters.subtract(hinters)
            self.update_busy_hints()

    @asyncSlot()
    async def update_streams(self, *args):
        """Update the list of available streams to choose from."""
        stream_names = await self.db.stream_names()
        # Sort so that "primary" is first
        sorted(stream_names, key=lambda x: x != "primary")
        self.ui.stream_combobox.clear()
        self.ui.stream_combobox.addItems(stream_names)

    @property
    def stream(self):
        current_text = self.ui.stream_combobox.currentText()
        return current_text or "primary"

    @asyncSlot()
    @cancellable
    async def update_multi_signals(self, *args):
        """Retrieve a new list of signals for multi plot and update UI."""
        combobox = self.ui.multi_signal_x_combobox
        # Store old value for restoring later
        old_value = combobox.currentText()
        # Determine valid list of columns to choose from
        use_hints = self.ui.plot_multi_hints_checkbox.isChecked()
        signals_task = self.db_task(
            self.db.signal_names(hinted_only=use_hints, stream=self.stream),
            "multi signals",
        )
        xcols, ycols = await signals_task
        # Update the comboboxes with new signals
        old_cols = [combobox.itemText(idx) for idx in range(combobox.count())]
        if xcols != old_cols:
            combobox.clear()
            combobox.addItems(xcols)
            # Restore previous value
            combobox.setCurrentText(old_value)

    @asyncSlot()
    @cancellable
    async def update_1d_signals(self, *args):
        # Store old values for restoring later
        comboboxes = [
            self.ui.signal_x_combobox,
            self.ui.signal_y_combobox,
            self.ui.signal_r_combobox,
        ]
        old_values = [cb.currentText() for cb in comboboxes]
        # Determine valid list of columns to choose from
        use_hints = self.ui.plot_1d_hints_checkbox.isChecked()
        signals_task = self.db_task(
            self.db.signal_names(hinted_only=use_hints, stream=self.stream),
            name="1D signals",
        )
        xcols, ycols = await signals_task
        self.multi_y_signals = ycols
        # Update the comboboxes with new signals
        self.ui.signal_x_combobox.clear()
        self.ui.signal_x_combobox.addItems(xcols)
        for cb in [
            self.ui.signal_y_combobox,
            self.ui.signal_r_combobox,
        ]:
            cb.clear()
            cb.addItems(ycols)
        # Restore previous values
        for val, cb in zip(old_values, comboboxes):
            cb.setCurrentText(val)

    @asyncSlot()
    @cancellable
    async def update_2d_signals(self, *args):
        # Store current selection for restoring later
        val_cb = self.ui.signal_value_combobox
        old_value = val_cb.currentText()
        # Determine valid list of dependent signals to choose from
        use_hints = self.ui.plot_2d_hints_checkbox.isChecked()
        xcols, vcols = await self.db_task(
            self.db.signal_names(hinted_only=use_hints, stream=self.stream),
            "2D signals",
        )
        # Update the UI with the list of controls
        val_cb.clear()
        val_cb.addItems(vcols)
        # Restore previous selection
        val_cb.setCurrentText(old_value)

    @asyncSlot()
    @cancellable
    async def update_multi_plot(self, *args):
        x_signal = self.ui.multi_signal_x_combobox.currentText()
        if x_signal == "":
            return
        use_hints = self.ui.plot_multi_hints_checkbox.isChecked()
        runs = await self.db_task(
            self.db.all_signals(hinted_only=use_hints, stream=self.stream), "multi-plot"
        )
        self.ui.plot_multi_view.plot_runs(runs, xsignal=x_signal)

    def update_export_button(self):
        # We can only export one scan at a time from here
        should_enable = self.selected_runs is not None and len(self.selected_runs) == 1
        self.ui.export_button.setEnabled(should_enable)

    @asyncSlot()
    async def export_runs(self):
        """Export the selected runs to user-specified filenames.

        Shows the user a file dialog with accepted types based on the
        accepted tiled export formats.

        """
        dialog = self.export_dialog
        # Determine default mimetypes
        mimetypes = self.selected_runs[0].formats()
        filenames = dialog.ask(mimetypes=mimetypes)
        mimetype = dialog.selectedMimeTypeFilter()
        formats = [mimetype] * len(filenames)
        await self.db_task(self.db.export_runs(filenames, formats=formats), "export")

    @asyncSlot(str)
    @cancellable
    async def update_running_scan(self, uid: str):
        log.debug(f"Updating running scan: {uid=}")
        await self.update_1d_plot(uids=[uid])

    @asyncSlot()
    @cancellable
    async def update_1d_plot(self, *args, uids: Sequence[str] = None):
        """Updates the data used in the plots.

        If *uids* is given, only runs with UIDs listed in *uids* will
        be updated.

        """
        # Figure out which signals to plot
        y_signal = self.ui.signal_y_combobox.currentText()
        x_signal = self.ui.signal_x_combobox.currentText()
        use_reference = self.ui.signal_r_checkbox.isChecked()
        if use_reference:
            r_signal = self.ui.signal_r_combobox.currentText()
        else:
            r_signal = None
        use_log = self.ui.logarithm_checkbox.isChecked()
        use_invert = self.ui.invert_checkbox.isChecked()
        use_grad = self.ui.gradient_checkbox.isChecked()
        # Load data
        task = self.db_task(
            self.db.signals(
                x_signal,
                y_signal,
                r_signal,
                use_log=use_log,
                use_invert=use_invert,
                use_grad=use_grad,
                uids=uids,
                stream=self.stream,
            ),
            "1D plot",
        )
        runs = await task
        if len(runs) == 0:
            return
        # Decide on axes labels
        xlabel = x_signal
        if r_signal is not None:
            if use_invert:
                ylabel = f"{r_signal}/{y_signal}"
            else:
                ylabel = f"{y_signal}/{r_signal}"
        else:
            if use_invert:
                ylabel = f"1/{y_signal}"
            else:
                ylabel = y_signal
        if use_log:
            ylabel = f"ln({ylabel})"
        if use_grad:
            ylabel = f"∇ {ylabel}"
        # Do the plotting
        self.ui.plot_1d_view.plot_runs(runs, xlabel=xlabel, ylabel=ylabel)
        if self.ui.autorange_1d_checkbox.isChecked():
            self.ui.plot_1d_view.autoRange()

    @asyncSlot()
    @cancellable
    async def update_2d_plot(self):
        """Change the 2D map plot based on desired signals, etc."""
        # Figure out which signals to plot
        value_signal = self.ui.signal_value_combobox.currentText()
        use_log = self.ui.logarithm_checkbox_2d.isChecked()
        use_invert = self.ui.invert_checkbox_2d.isChecked()
        use_grad = self.ui.gradient_checkbox_2d.isChecked()
        images = await self.db_task(
            self.db.images(value_signal, stream=self.stream), "2D plot"
        )
        # Get axis labels
        # Eventually this will be replaced with robust choices for plotting multiple images
        metadata = await self.db_task(self.db.metadata(), "2D plot")
        metadata = list(metadata.values())[0]
        dimensions = metadata["start"]["hints"]["dimensions"]
        try:
            xlabel = dimensions[-1][0][0]
            ylabel = dimensions[-2][0][0]
        except IndexError:
            # Not a 2D scan
            return
        # Get spatial extent
        extents = metadata["start"]["extents"]
        self.ui.plot_2d_view.plot_runs(
            images, xlabel=xlabel, ylabel=ylabel, extents=extents
        )

    @asyncSlot()
    async def update_metadata(self, *args):
        """Render metadata for the runs into the metadata widget."""
        # Combine the metadata in a human-readable output
        text = ""
        all_md = await self.db_task(self.db.metadata(), "metadata")
        for uid, md in all_md.items():
            text += f"# {uid}"
            text += yaml.dump(md)
            text += f"\n\n{'=' * 20}\n\n"
        # Update the widget with the rendered metadata
        self.ui.metadata_textedit.document().setPlainText(text)

    def clear_plots(self):
        """Clear all the plots.

        If a *uid* is provided, only the plots matching the scan with
        *uid* will be updated.
        """
        self.plot_1d_view.clear_runs()

    @asyncSlot()
    @cancellable
    async def update_plots(self):
        """Get new data, and update all the plots.

        If a *uid* is provided, only the plots matching the scan with
        *uid* will be updated.
        """

        await asyncio.gather(
            self.update_metadata(),
            self.update_1d_plot(),
            self.update_2d_plot(),
            self.update_multi_plot(),
        )

    @asyncSlot()
    @cancellable
    async def update_selected_runs(self, *args):
        """Get the current runs from the database and stash them."""
        # Get UID's from the selection
        col_idx = self._run_col_names.index("UID")
        indexes = self.ui.run_tableview.selectedIndexes()
        uids = [i.siblingAtColumn(col_idx).data() for i in indexes]
        # Get selected runs from the database
        with self.busy_hints(run_widgets=True, run_table=False, filter_widgets=False):
            task = self.db_task(
                self.db.load_selected_runs(uids=uids), "update selected runs"
            )
            self.selected_runs = await task
            # Update the necessary UI elements
            await asyncio.gather(
                self.update_multi_signals(),
                self.update_1d_signals(),
                self.update_2d_signals(),
                self.update_streams(),
            )
            # Update the plots
            self.clear_plots()
            await self.update_plots()
            self.update_export_button()

    def filters(self, *args):
        new_filters = {
            "plan": self.ui.filter_plan_combobox.currentText(),
            "sample": self.ui.filter_sample_combobox.currentText(),
            "formula": self.ui.filter_formula_combobox.currentText(),
            "edge": self.ui.filter_edge_combobox.currentText(),
            "exit_status": self.ui.filter_exit_status_combobox.currentText(),
            "user": self.ui.filter_user_combobox.currentText(),
            "proposal": self.ui.filter_proposal_combobox.currentText(),
            "esaf": self.ui.filter_esaf_combobox.currentText(),
            "beamline": self.ui.filter_beamline_combobox.currentText(),
            "full_text": self.ui.filter_full_text_lineedit.text(),
        }
        # Special handling for the time-based filters
        if self.ui.filter_after_checkbox.checkState():
            after = self.ui.filter_after_datetimeedit.dateTime().toSecsSinceEpoch()
            new_filters["after"] = after
        if self.ui.filter_before_checkbox.checkState():
            before = self.ui.filter_before_datetimeedit.dateTime().toSecsSinceEpoch()
            new_filters["before"] = before
        # Limit the search to standards only
        if self.ui.filter_standards_checkbox.checkState():
            new_filters["standards_only"] = True
        # Only include values that were actually filled in
        null_values = ["", False]
        new_filters = {k: v for k, v in new_filters.items() if v not in null_values}
        return new_filters

    def load_models(self):
        # Set up the model
        self.runs_model = QStandardItemModel()
        # Add the model to the UI element
        self.ui.run_tableview.setModel(self.runs_model)

    def ui_filename(self):
        return "run_browser/run_browser.ui"



import logging
import re  # noqa: F401

import databroker  # noqa: F401
from bluesky.plan_stubs import abs_set  # noqa: F401
from bluesky.plan_stubs import mv as _mv  # noqa: F401
from bluesky.plan_stubs import mvr, null, pause, rel_set, sleep, stop  # noqa: F401
from bluesky.run_engine import call_in_bluesky_event_loop
from ophyd_async.core import NotConnected

# Import plans
from haven import beamline, recall_motor_position, sanitize_name  # noqa: F401
from haven.plans import (  # noqa: F401
    auto_gain,
    count,
    energy_scan,
    grid_scan,
    list_scan,
    record_dark_current,
    rel_grid_scan,
    rel_list_scan,
    rel_scan,
    scan,
    scan_nd,
    set_energy,
    xafs_scan,
)
from haven.run_engine import run_engine  # noqa: F401

log = logging.getLogger(__name__)

# Create a run engine
RE = run_engine(
    connect_databroker=False, connect_tiled=False, connect_kafka=True, use_bec=False
)

# Import devices
beamline.load()
try:
    call_in_bluesky_event_loop(beamline.connect())
except NotConnected as exc:
    log.exception(exc)
for cpt in beamline.devices._objects_by_name.values():
    # Replace spaces and other illegal characters in variable name
    # name = re.sub('\W|^(?=\d)','_', cpt.name)
    name = sanitize_name(cpt.name)
    # Add the device as a variable in module's globals
    globals().setdefault(name, cpt)


# Workaround for https://github.com/bluesky/bluesky-queueserver/issues/310
from collections.abc import Hashable


def mv(
    *args,
    group: Hashable | None = None,
    **kwargs,
):
    """
    Move one or more devices to a setpoint. Wait for all to complete.

    If more than one device is specified, the movements are done in parallel.

    Parameters
    ----------
    args :
        device1, value1, device2, value2, ...
    group : string, optional
        Used to mark these as a unit to be waited on.
    kwargs :
        passed to obj.set()

    Yields
    ------
    msg : Msg

    Returns
    -------
    statuses :
        Tuple of n statuses, one for each move operation

    See Also
    --------
    :func:`bluesky.plan_stubs.abs_set`
    :func:`bluesky.plan_stubs.mvr`
    """
    yield from _mv(*args, group=group, **kwargs)



import asyncio

from ophyd_async.core import (
    Device,
    StandardReadable,
    StandardReadableFormat,
    StrictEnum,
    SubsetEnum,
)
from ophyd_async.epics.core import epics_signal_r, epics_signal_rw, epics_signal_x


class AlarmStatus(SubsetEnum):
    NO_ALARM = "NO_ALARM"
    READ = "READ"
    WRITE = "WRITE"
    HIHI = "HIHI"
    HIGH = "HIGH"
    LOLO = "LOLO"
    LOW = "LOW"
    STATE = "STATE"
    COS = "COS"
    COMM = "COMM"
    TIMEOUT = "TIMEOUT"
    HWLIMIT = "HWLIMIT"
    CALC = "CALC"
    SCAN = "SCAN"
    LINK = "LINK"
    SOFT = "SOFT"
    # BAD_SUB = "BAD_SUB"
    # UDF = "UDF"
    # DISABLE = "DISABLE"
    # SIMM = "SIMM"
    # READ_ACCESS = "READ_ACCESS"
    # WRITE_ACCESS = "WRITE_ACCESS"


class AlarmSeverity(StrictEnum):
    NO_ALARM = "NO_ALARM"
    MINOR = "MINOR"
    MAJOR = "MAJOR"
    INVALID = "INVALID"


class EpicsRecordDeviceCommonAll(StandardReadable):
    """
    Many of the fields common to all EPICS records.

    Some fields are not included because they are not interesting to
    an EPICS client or are already provided in other support.
    """

    # More valid options are specific to the record type
    # Subclasses may override this attribute
    class ScanInterval(SubsetEnum):
        PASSIVE = "Passive"
        EVENT = "Event"
        IO_INTR = "I/O Intr"
        SCAN_10 = "10 second"
        SCAN_5 = "5 second"
        SCAN_2 = "2 second"
        SCAN_1 = "1 second"
        SCAN_0_5 = ".5 second"
        SCAN_0_2 = ".2 second"
        SCAN_0_1 = ".1 second"

    # Config signals
    def __init__(self, prefix: str, name: str = ""):
        with self.add_children_as_readables(StandardReadableFormat.CONFIG_SIGNAL):
            self.description = epics_signal_rw(str, f"{prefix}.DESC")
            self.scanning_rate = epics_signal_rw(self.ScanInterval, f"{prefix}.SCAN")
        # Other signals, not included in read
        self.disable_value = epics_signal_rw(int, f"{prefix}.DISV")
        self.scan_disable_input_link_value = epics_signal_rw(int, f"{prefix}.DISA")
        self.scan_disable_value_input_link = epics_signal_rw(str, f"{prefix}.SDIS")
        self.forward_link = epics_signal_rw(str, f"{prefix}.FLNK")

        self.alarm_status = epics_signal_r(AlarmStatus, f"{prefix}.STAT")
        self.alarm_severity = epics_signal_r(AlarmSeverity, f"{prefix}.SEVR")
        self.new_alarm_status = epics_signal_r(AlarmStatus, f"{prefix}.NSTA")
        self.new_alarm_severity = epics_signal_r(AlarmSeverity, f"{prefix}.NSEV")
        self.disable_alarm_severity = epics_signal_rw(AlarmSeverity, f"{prefix}.DISS")
        self.processing_active = epics_signal_r(int, f"{prefix}.PACT")
        self.process_record = epics_signal_x(f"{prefix}.PROC")
        self.trace_processing = epics_signal_rw(int, f"{prefix}.TPRO")

        super().__init__(name=name)


class EpicsSynAppsRecordEnableMixin(Device):
    """Supports ``{PV}Enable`` feature from user databases."""

    def __init__(self, prefix, name=""):
        with self.add_children_as_readables(StandardReadableFormat.CONFIG_SIGNAL):
            self.enable = epics_signal_rw(int, "Enable")
        super().__init__(name=name)

    async def reset(self):
        """set all fields to default values"""
        await asyncio.gather(
            self.enable.set(self.enable.enum_strs[1]), super().reset()  # Enable
        )


class EpicsRecordInputFields(EpicsRecordDeviceCommonAll):
    """
    Some fields common to EPICS input records.
    """

    class DeviceType(SubsetEnum):
        SOFT_CHANNEL = "Soft Channel"
        RAW_SOFT_CHANNEL = "Raw Soft Channel"
        ASYNC_SOFT_CHANNEL = "Async Soft Channel"

    def __init__(self, prefix: str, name: str = ""):
        with self.add_children_as_readables(StandardReadableFormat.CONFIG_SIGNAL):
            self.input_link = epics_signal_rw(str, f"{prefix}.INP")
            self.device_type = epics_signal_r(self.DeviceType, f"{prefix}.DTYP")
        super().__init__(prefix=prefix, name=name)


class EpicsRecordOutputFields(EpicsRecordDeviceCommonAll):
    """
    Some fields common to EPICS output records.
    """

    class DeviceType(SubsetEnum):
        SOFT_CHANNEL = "Soft Channel"
        RAW_SOFT_CHANNEL = "Raw Soft Channel"
        ASYNC_SOFT_CHANNEL = "Async Soft Channel"

    class ModeSelect(SubsetEnum):
        SUPERVISORY = "supervisory"
        CLOSED_LOOP = "closed_loop"

    def __init__(self, prefix: str, name: str = ""):
        with self.add_children_as_readables(StandardReadableFormat.CONFIG_SIGNAL):
            self.output_link = epics_signal_rw(str, f"{prefix}.OUT")
            self.desired_output_location = epics_signal_rw(str, f"{prefix}.DOL")
            self.output_mode_select = epics_signal_rw(self.ModeSelect, f"{prefix}.OMSL")
            self.device_type = epics_signal_r(self.DeviceType, f"{prefix}.DTYP")
        super().__init__(prefix=prefix, name=name)



from unittest import mock

import pytest

from firefly.kafka_client import KafkaClient


@pytest.fixture()
def client():
    consumer = mock.AsyncMock()
    return KafkaClient(kafka_consumer=consumer)


async def test_start(client):
    await client.consumer_loop()


async def test_start_twice(client):
    client.start()
    with pytest.raises(Exception):
        client.start()


def check_uid(uid):
    return uid == "40b434c5-bec2-4eff-9cbe-c470b6435677"


def test_start_message(client, qtbot):
    with qtbot.waitSignal(client.run_started, check_params_cb=check_uid):
        client._process_document("start", start_doc)


def test_event_message(client, qtbot):
    with qtbot.waitSignal(client.run_updated, check_params_cb=check_uid):
        client._process_document("descriptor", descriptor_doc)
        client._process_document("event", event_doc)


def test_stop_message(client, qtbot):
    client._process_document("descriptor", descriptor_doc)
    assert len(client._descriptors) == 1
    with qtbot.waitSignal(client.run_stopped, check_params_cb=check_uid):

        client._process_document("stop", stop_doc)
    assert len(client._descriptors) == 0


start_doc = {
    "EPICS_CA_MAX_ARRAY_BYTES": "16777216",
    "EPICS_HOST_ARCH": "linux-x86_64",
    "beamline_id": "25-ID-C",
    "detectors": ["I0"],
    "epics_libca": "/home/beams0/S25STAFF/miniforge3/envs/haven-dev/epics/lib/linux-x86_64/libca.so",
    "facility_id": "Advanced Photon Source",
    "hints": {"dimensions": [[["time"], "primary"]]},
    "login_id": "s25staff@fedorov.xray.aps.anl.gov",
    "num_intervals": 0,
    "num_points": 1,
    "parameters": "",
    "pid": 1080697,
    "plan_args": {
        "delay": None,
        "detectors": [
            "IonChamber(prefix='25idcVME:3820:', name='I0', "
            "read_attrs=['voltmeter', 'voltmeter.enable', "
            "'voltmeter.amps', 'voltmeter.volts', 'volts', "
            "'counts', 'net_counts', 'exposure_time'], "
            "configuration_attrs=['description', "
            "'voltmeter', 'voltmeter.description', "
            "'voltmeter.scanning_rate', "
            "'voltmeter.disable_value', "
            "'voltmeter.scan_disable_input_link_value', "
            "'voltmeter.scan_disable_value_input_link', "
            "'voltmeter.forward_link', "
            "'voltmeter.device_type', "
            "'voltmeter.alarm_status', "
            "'voltmeter.alarm_severity', "
            "'voltmeter.new_alarm_status', "
            "'voltmeter.new_alarm_severity', "
            "'voltmeter.disable_alarm_severity', "
            "'voltmeter.input_link', 'voltmeter.raw_value', "
            "'voltmeter.differential', 'voltmeter.high', "
            "'voltmeter.low', 'voltmeter.temperature_units', "
            "'voltmeter.resolution', 'voltmeter.range', "
            "'voltmeter.mode', 'gate', 'preset_count', "
            "'frequency', 'offset', 'record_dark_time', "
            "'channel_advance_source', "
            "'num_channels_to_use', 'max_channels', "
            "'channel_one_source', 'count_on_start'])"
        ],
        "num": 1,
    },
    "plan_name": "count",
    "plan_type": "generator",
    "purpose": "",
    "sample_name": "",
    "scan_id": 2,
    "time": 1717042074.4366016,
    "uid": "40b434c5-bec2-4eff-9cbe-c470b6435677",
    "versions": {
        "apstools": "1.6.19.dev52+ga165d24",
        "bluesky": "1.13.0a3",
        "databroker": "1.2.5",
        "epics": "3.5.2",
        "epics_ca": "3.5.2",
        "h5py": "3.9.0",
        "haven": "23.10.0",
        "matplotlib": "3.8.4",
        "numpy": "1.26.4",
        "ophyd": "1.9.0",
        "pymongo": "4.7.2",
    },
    "xray_source": "insertion device",
}
descriptor_doc = {
    "configuration": {
        "I0": {
            "data": {
                "I0_channel_advance_source": 1,
                "I0_channel_one_source": 1,
                "I0_count_on_start": 1,
                "I0_description": "I0",
                "I0_frequency": 10000000.0,
                "I0_gate": 0,
                "I0_max_channels": 8000,
                "I0_num_channels_to_use": 8000,
                "I0_offset": 0.0,
                "I0_preset_count": 0.0,
                "I0_record_dark_time": 0.0,
                "I0_voltmeter_alarm_severity": 0,
                "I0_voltmeter_alarm_status": 0,
                "I0_voltmeter_description": "I0",
                "I0_voltmeter_device_type": 8,
                "I0_voltmeter_differential": 0,
                "I0_voltmeter_disable_alarm_severity": 0,
                "I0_voltmeter_disable_value": 1,
                "I0_voltmeter_forward_link": "",
                "I0_voltmeter_high": 0.0,
                "I0_voltmeter_input_link": "@asyn(LJT7V_0 " "2)ANALOG_IN_VALUE",
                "I0_voltmeter_low": 0.0,
                "I0_voltmeter_mode": 0,
                "I0_voltmeter_new_alarm_severity": 0,
                "I0_voltmeter_new_alarm_status": 0,
                "I0_voltmeter_range": 0,
                "I0_voltmeter_raw_value": 0,
                "I0_voltmeter_resolution": 0,
                "I0_voltmeter_scan_disable_input_link_value": 0,
                "I0_voltmeter_scan_disable_value_input_link": "",
                "I0_voltmeter_scanning_rate": 9,
                "I0_voltmeter_temperature_units": 0,
            },
            "data_keys": {
                "I0_channel_advance_source": {
                    "dtype": "integer",
                    "enum_strs": ["Internal", "External"],
                    "lower_ctrl_limit": None,
                    "shape": [],
                    "source": "PV:25idcVME:3820:ChannelAdvance",
                    "units": None,
                    "upper_ctrl_limit": None,
                },
                "I0_channel_one_source": {
                    "dtype": "integer",
                    "enum_strs": ["Int. " "clock", "External"],
                    "lower_ctrl_limit": None,
                    "shape": [],
                    "source": "PV:25idcVME:3820:Channel1Source",
                    "units": None,
                    "upper_ctrl_limit": None,
                },
                "I0_count_on_start": {
                    "dtype": "integer",
                    "enum_strs": ["No", "Yes"],
                    "lower_ctrl_limit": None,
                    "shape": [],
                    "source": "PV:25idcVME:3820:CountOnStart",
                    "units": None,
                    "upper_ctrl_limit": None,
                },
                "I0_description": {
                    "dtype": "string",
                    "lower_ctrl_limit": None,
                    "shape": [],
                    "source": "PV:25idcVME:3820:scaler1.NM4",
                    "units": None,
                    "upper_ctrl_limit": None,
                },
                "I0_frequency": {
                    "dtype": "number",
                    "lower_ctrl_limit": 0.0,
                    "precision": 3,
                    "shape": [],
                    "source": "PV:25idcVME:3820:scaler1.FREQ",
                    "units": "",
                    "upper_ctrl_limit": 0.0,
                },
                "I0_gate": {
                    "dtype": "integer",
                    "enum_strs": ["N", "Y"],
                    "lower_ctrl_limit": None,
                    "shape": [],
                    "source": "PV:25idcVME:3820:scaler1.G4",
                    "units": None,
                    "upper_ctrl_limit": None,
                },
                "I0_max_channels": {
                    "dtype": "integer",
                    "lower_ctrl_limit": 0,
                    "shape": [],
                    "source": "PV:25idcVME:3820:MaxChannels",
                    "units": "",
                    "upper_ctrl_limit": 0,
                },
                "I0_num_channels_to_use": {
                    "dtype": "integer",
                    "lower_ctrl_limit": 0,
                    "shape": [],
                    "source": "PV:25idcVME:3820:NuseAll",
                    "units": "",
                    "upper_ctrl_limit": 0,
                },
                "I0_offset": {
                    "dtype": "number",
                    "lower_ctrl_limit": 0.0,
                    "precision": 0,
                    "shape": [],
                    "source": "PV:25idcVME:3820:scaler1_offset0.D",
                    "units": "",
                    "upper_ctrl_limit": 0.0,
                },
                "I0_preset_count": {
                    "dtype": "number",
                    "lower_ctrl_limit": 0.0,
                    "precision": 0,
                    "shape": [],
                    "source": "PV:25idcVME:3820:scaler1.PR4",
                    "units": "",
                    "upper_ctrl_limit": 0.0,
                },
                "I0_record_dark_time": {
                    "dtype": "number",
                    "lower_ctrl_limit": 0.0,
                    "precision": 0,
                    "shape": [],
                    "source": "PV:25idcVME:3820:scaler1_offset_time.VAL",
                    "units": "",
                    "upper_ctrl_limit": 0.0,
                },
                "I0_voltmeter_alarm_severity": {
                    "dtype": "integer",
                    "enum_strs": [
                        "NO_ALARM",
                        "MINOR",
                        "MAJOR",
                        "INVALID",
                    ],
                    "lower_ctrl_limit": None,
                    "shape": [],
                    "source": "PV:25idc:LJT7Voltmeter_0:Ai2.SEVR",
                    "units": None,
                    "upper_ctrl_limit": None,
                },
                "I0_voltmeter_alarm_status": {
                    "dtype": "integer",
                    "enum_strs": [
                        "NO_ALARM",
                        "READ",
                        "WRITE",
                        "HIHI",
                        "HIGH",
                        "LOLO",
                        "LOW",
                        "STATE",
                        "COS",
                        "COMM",
                        "TIMEOUT",
                        "HWLIMIT",
                        "CALC",
                        "SCAN",
                        "LINK",
                        "SOFT",
                    ],
                    "lower_ctrl_limit": None,
                    "shape": [],
                    "source": "PV:25idc:LJT7Voltmeter_0:Ai2.STAT",
                    "units": None,
                    "upper_ctrl_limit": None,
                },
                "I0_voltmeter_description": {
                    "dtype": "string",
                    "lower_ctrl_limit": None,
                    "shape": [],
                    "source": "PV:25idc:LJT7Voltmeter_0:Ai2.DESC",
                    "units": None,
                    "upper_ctrl_limit": None,
                },
                "I0_voltmeter_device_type": {
                    "dtype": "integer",
                    "enum_strs": [
                        "Soft " "Channel",
                        "Raw " "Soft " "Channel",
                        "Async " "Soft " "Channel",
                        "Soft " "Timestamp",
                        "General " "Time",
                        "asynInt32",
                        "asynInt32Average",
                        "asynFloat64",
                        "asynFloat64Average",
                        "asynInt64",
                        "IOC " "stats",
                        "IOC " "stats " "clusts",
                        "GPIB " "init/report",
                        "Sec " "Past " "Epoch",
                        "asyn " "ai " "stringParm",
                        "asyn " "ai " "HeidND261",
                    ],
                    "lower_ctrl_limit": None,
                    "shape": [],
                    "source": "PV:25idc:LJT7Voltmeter_0:Ai2.DTYP",
                    "units": None,
                    "upper_ctrl_limit": None,
                },
                "I0_voltmeter_differential": {
                    "dtype": "integer",
                    "enum_strs": ["Single-Ended", "Differential"],
                    "lower_ctrl_limit": None,
                    "shape": [],
                    "source": "PV:25idc:LJT7Voltmeter_0:AiDiff2",
                    "units": None,
                    "upper_ctrl_limit": None,
                },
                "I0_voltmeter_disable_alarm_severity": {
                    "dtype": "integer",
                    "enum_strs": [
                        "NO_ALARM",
                        "MINOR",
                        "MAJOR",
                        "INVALID",
                    ],
                    "lower_ctrl_limit": None,
                    "shape": [],
                    "source": "PV:25idc:LJT7Voltmeter_0:Ai2.DISS",
                    "units": None,
                    "upper_ctrl_limit": None,
                },
                "I0_voltmeter_disable_value": {
                    "dtype": "integer",
                    "lower_ctrl_limit": -32768,
                    "shape": [],
                    "source": "PV:25idc:LJT7Voltmeter_0:Ai2.DISV",
                    "units": "",
                    "upper_ctrl_limit": 32767,
                },
                "I0_voltmeter_forward_link": {
                    "dtype": "string",
                    "lower_ctrl_limit": None,
                    "shape": [],
                    "source": "PV:25idc:LJT7Voltmeter_0:Ai2.FLNK",
                    "units": None,
                    "upper_ctrl_limit": None,
                },
                "I0_voltmeter_high": {
                    "dtype": "number",
                    "lower_ctrl_limit": 0.0,
                    "precision": 4,
                    "shape": [],
                    "source": "PV:25idc:LJT7Voltmeter_0:AiHOPR2",
                    "units": "",
                    "upper_ctrl_limit": 0.0,
                },
                "I0_voltmeter_input_link": {
                    "dtype": "string",
                    "lower_ctrl_limit": None,
                    "shape": [],
                    "source": "PV:25idc:LJT7Voltmeter_0:Ai2.INP",
                    "units": None,
                    "upper_ctrl_limit": None,
                },
                "I0_voltmeter_low": {
                    "dtype": "number",
                    "lower_ctrl_limit": 0.0,
                    "precision": 4,
                    "shape": [],
                    "source": "PV:25idc:LJT7Voltmeter_0:AiLOPR2",
                    "units": "",
                    "upper_ctrl_limit": 0.0,
                },
                "I0_voltmeter_mode": {
                    "dtype": "integer",
                    "enum_strs": [
                        "Volts",
                        "Type " "B " "TC",
                        "Type " "C " "TC",
                        "Type " "E " "TC",
                        "Type " "J " "TC",
                        "Type " "K " "TC",
                        "Type " "N " "TC",
                        "Type " "R " "TC",
                        "Type " "S " "TC",
                        "Type " "T " "TC",
                    ],
                    "lower_ctrl_limit": None,
                    "shape": [],
                    "source": "PV:25idc:LJT7Voltmeter_0:AiMode2",
                    "units": None,
                    "upper_ctrl_limit": None,
                },
                "I0_voltmeter_new_alarm_severity": {
                    "dtype": "integer",
                    "enum_strs": [
                        "NO_ALARM",
                        "MINOR",
                        "MAJOR",
                        "INVALID",
                    ],
                    "lower_ctrl_limit": None,
                    "shape": [],
                    "source": "PV:25idc:LJT7Voltmeter_0:Ai2.NSEV",
                    "units": None,
                    "upper_ctrl_limit": None,
                },
                "I0_voltmeter_new_alarm_status": {
                    "dtype": "integer",
                    "enum_strs": [
                        "NO_ALARM",
                        "READ",
                        "WRITE",
                        "HIHI",
                        "HIGH",
                        "LOLO",
                        "LOW",
                        "STATE",
                        "COS",
                        "COMM",
                        "TIMEOUT",
                        "HWLIMIT",
                        "CALC",
                        "SCAN",
                        "LINK",
                        "SOFT",
                    ],
                    "lower_ctrl_limit": None,
                    "shape": [],
                    "source": "PV:25idc:LJT7Voltmeter_0:Ai2.NSTA",
                    "units": None,
                    "upper_ctrl_limit": None,
                },
                "I0_voltmeter_range": {
                    "dtype": "integer",
                    "enum_strs": [
                        "+= " "10V",
                        "+= " "1V",
                        "+= " "0.1V",
                        "+= " "0.01V",
                    ],
                    "lower_ctrl_limit": None,
                    "shape": [],
                    "source": "PV:25idc:LJT7Voltmeter_0:AiRange2",
                    "units": None,
                    "upper_ctrl_limit": None,
                },
                "I0_voltmeter_raw_value": {
                    "dtype": "integer",
                    "lower_ctrl_limit": -2147483648,
                    "shape": [],
                    "source": "PV:25idc:LJT7Voltmeter_0:Ai2.RVAL",
                    "units": "",
                    "upper_ctrl_limit": 2147483647,
                },
                "I0_voltmeter_resolution": {
                    "dtype": "integer",
                    "enum_strs": [
                        "Default",
                        "1",
                        "2",
                        "3",
                        "4",
                        "5",
                        "6",
                        "7",
                        "8",
                    ],
                    "lower_ctrl_limit": None,
                    "shape": [],
                    "source": "PV:25idc:LJT7Voltmeter_0:AiResolution2",
                    "units": None,
                    "upper_ctrl_limit": None,
                },
                "I0_voltmeter_scan_disable_input_link_value": {
                    "dtype": "integer",
                    "lower_ctrl_limit": -32768,
                    "shape": [],
                    "source": "PV:25idc:LJT7Voltmeter_0:Ai2.DISA",
                    "units": "",
                    "upper_ctrl_limit": 32767,
                },
                "I0_voltmeter_scan_disable_value_input_link": {
                    "dtype": "string",
                    "lower_ctrl_limit": None,
                    "shape": [],
                    "source": "PV:25idc:LJT7Voltmeter_0:Ai2.SDIS",
                    "units": None,
                    "upper_ctrl_limit": None,
                },
                "I0_voltmeter_scanning_rate": {
                    "dtype": "integer",
                    "enum_strs": [
                        "Passive",
                        "Event",
                        "I/O " "Intr",
                        "10 " "second",
                        "5 " "second",
                        "2 " "second",
                        "1 " "second",
                        ".5 " "second",
                        ".2 " "second",
                        ".1 " "second",
                    ],
                    "lower_ctrl_limit": None,
                    "shape": [],
                    "source": "PV:25idc:LJT7Voltmeter_0:Ai2.SCAN",
                    "units": None,
                    "upper_ctrl_limit": None,
                },
                "I0_voltmeter_temperature_units": {
                    "dtype": "integer",
                    "enum_strs": ["K", "C", "F"],
                    "lower_ctrl_limit": None,
                    "shape": [],
                    "source": "PV:25idc:LJT7Voltmeter_0:AiTempUnits2",
                    "units": None,
                    "upper_ctrl_limit": None,
                },
            },
            "timestamps": {
                "I0_channel_advance_source": 1706288136.019154,
                "I0_channel_one_source": 1715890566.204044,
                "I0_count_on_start": 1716408876.514808,
                "I0_description": 1716500670.770429,
                "I0_frequency": 1716500670.770429,
                "I0_gate": 1716500670.770429,
                "I0_max_channels": 1706288136.002487,
                "I0_num_channels_to_use": 1706288136.002487,
                "I0_offset": 631152000.0,
                "I0_preset_count": 1716500670.770429,
                "I0_record_dark_time": 631152000.0,
                "I0_voltmeter_alarm_severity": 1717042054.125238,
                "I0_voltmeter_alarm_status": 1717042054.125238,
                "I0_voltmeter_description": 1717041575.425234,
                "I0_voltmeter_device_type": 1717041573.025238,
                "I0_voltmeter_differential": 1715889930.536937,
                "I0_voltmeter_disable_alarm_severity": 1717041573.025238,
                "I0_voltmeter_disable_value": 1717041573.025238,
                "I0_voltmeter_forward_link": 1717041575.425234,
                "I0_voltmeter_high": 631152000.0,
                "I0_voltmeter_input_link": 1717041575.425234,
                "I0_voltmeter_low": 631152000.0,
                "I0_voltmeter_mode": 1715889930.536705,
                "I0_voltmeter_new_alarm_severity": 1717041573.025238,
                "I0_voltmeter_new_alarm_status": 1717041573.025238,
                "I0_voltmeter_range": 1715889930.537142,
                "I0_voltmeter_raw_value": 1717041573.025238,
                "I0_voltmeter_resolution": 1715889930.537357,
                "I0_voltmeter_scan_disable_input_link_value": 1717041573.025238,
                "I0_voltmeter_scan_disable_value_input_link": 1717041575.425234,
                "I0_voltmeter_scanning_rate": 1717041573.025238,
                "I0_voltmeter_temperature_units": 1715889930.536706,
            },
        }
    },
    "data_keys": {
        "I0_counts": {
            "dtype": "number",
            "lower_ctrl_limit": 0.0,
            "object_name": "I0",
            "precision": 0,
            "shape": [],
            "source": "PV:25idcVME:3820:scaler1.S4",
            "units": "",
            "upper_ctrl_limit": 0.0,
        },
        "I0_exposure_time": {
            "dtype": "number",
            "lower_ctrl_limit": 0.0,
            "object_name": "I0",
            "precision": 3,
            "shape": [],
            "source": "PV:25idcVME:3820:scaler1.TP",
            "units": "",
            "upper_ctrl_limit": 0.0,
        },
        "I0_net_counts": {
            "dtype": "number",
            "lower_ctrl_limit": 0.0,
            "object_name": "I0",
            "precision": 0,
            "shape": [],
            "source": "PV:25idcVME:3820:scaler1_netA.D",
            "units": "",
            "upper_ctrl_limit": 0.0,
        },
        "I0_voltmeter_amps": {
            "derived_from": "I0_voltmeter_volts",
            "dtype": "number",
            "lower_ctrl_limit": 0.0,
            "object_name": "I0",
            "precision": 4,
            "shape": [],
            "source": "SIM:I0_voltmeter_amps",
            "units": "",
            "upper_ctrl_limit": 0.0,
        },
        "I0_voltmeter_enable": {
            "dtype": "integer",
            "enum_strs": ["Disable", "Enable"],
            "lower_ctrl_limit": None,
            "object_name": "I0",
            "shape": [],
            "source": "PV:25idc:LJT7Voltmeter_0:AiEnable2",
            "units": None,
            "upper_ctrl_limit": None,
        },
        "I0_voltmeter_volts": {
            "dtype": "number",
            "lower_ctrl_limit": 0.0,
            "object_name": "I0",
            "precision": 4,
            "shape": [],
            "source": "PV:25idc:LJT7Voltmeter_0:Ai2.VAL",
            "units": "",
            "upper_ctrl_limit": 0.0,
        },
        "I0_volts": {
            "derived_from": "I0_counts",
            "dtype": "number",
            "lower_ctrl_limit": 0.0,
            "object_name": "I0",
            "precision": 0,
            "shape": [],
            "source": "SIM:I0_volts",
            "units": "",
            "upper_ctrl_limit": 0.0,
        },
    },
    "hints": {"I0": {"fields": ["I0_net_counts"]}},
    "name": "primary",
    "object_keys": {
        "I0": [
            "I0_voltmeter_enable",
            "I0_voltmeter_amps",
            "I0_voltmeter_volts",
            "I0_volts",
            "I0_counts",
            "I0_net_counts",
            "I0_exposure_time",
        ]
    },
    "run_start": "40b434c5-bec2-4eff-9cbe-c470b6435677",
    "time": 1717042074.559484,
    "uid": "bb592c67-1d8f-41f1-99fa-d4db6886ae03",
}


event_doc = {
    "data": {
        "I0_counts": 989354.0,
        "I0_exposure_time": 0.1,
        "I0_net_counts": 989354.0,
        "I0_voltmeter_amps": -5.713760852813812e-10,
        "I0_voltmeter_enable": 1,
        "I0_voltmeter_volts": 0.9942862391471863,
        "I0_volts": 9.893530106469894,
    },
    "descriptor": "bb592c67-1d8f-41f1-99fa-d4db6886ae03",
    "filled": {},
    "seq_num": 1,
    "time": 1717042074.5687447,
    "timestamps": {
        "I0_counts": 1717042074.414631,
        "I0_exposure_time": 1716500670.770429,
        "I0_net_counts": 631152000.0,
        "I0_voltmeter_amps": 1717042074.52525,
        "I0_voltmeter_enable": 1715889930.53229,
        "I0_voltmeter_volts": 1717042074.52525,
        "I0_volts": 1717042074.414631,
    },
    "uid": "3577aa5f-0605-4609-af60-e5f9a9bc462e",
}


stop_doc = {
    "exit_status": "success",
    "num_events": {"primary": 1},
    "reason": "",
    "run_start": "40b434c5-bec2-4eff-9cbe-c470b6435677",
    "time": 1717042074.5689151,
    "uid": "e9a17152-2e39-4f1b-acab-71aeb11e7222",
}



"""Loader for creating instances of the devices from a config file."""

import logging
import os
from typing import Mapping

from guarneri import Instrument

from .devices.aerotech import AerotechStage
from .devices.aps import ApsMachine
from .devices.area_detector import make_area_detector
from .devices.beamline_manager import BeamlineManager
from .devices.detectors.aravis import AravisDetector
from .devices.detectors.sim_detector import SimDetector
from .devices.detectors.xspress import Xspress3Detector
from .devices.dxp import make_dxp_device
from .devices.energy_positioner import EnergyPositioner
from .devices.heater import CapillaryHeater
from .devices.ion_chamber import IonChamber
from .devices.mirrors import HighHeatLoadMirror, KBMirrors
from .devices.motor import Motor, load_motors
from .devices.power_supply import NHQ203MChannel
from .devices.robot import Robot
from .devices.scaler import MultiChannelScaler
from .devices.shutter import PssShutter
from .devices.slits import ApertureSlits, BladeSlits
from .devices.stage import XYStage
from .devices.table import Table
from .devices.xia_pfcu import PFCUFilterBank

log = logging.getLogger(__name__)


class HavenInstrument(Instrument):
    def load(
        self,
        config: Mapping = None,
        return_devices: bool = False,
        reset_devices: bool = True,
    ):
        """Load the beamline instrumentation.

        This function will reach out and query various IOCs for motor
        information based on the information in *config* (see
        ``iconfig_default.toml`` for examples). Based on the
        configuration, it will create Ophyd devices and register them with
        *registry*.

        Parameters
        ==========
        config:
          The beamline configuration read in from TOML files. Mostly
          useful for testing.
        return_devices
          If true, return the newly loaded devices when complete.
        reset_registry
          If true, existing devices will be removed before loading.

        """
        if reset_devices:
            self.devices.clear()
        # Check if config files are available
        if "HAVEN_CONFIG_FILES" in os.environ:
            config_files = os.environ.get("HAVEN_CONFIG_FILES", "").split(":")
        else:
            config_files = []
        # Load devices ("motors" is done later)
        for cfg_file in config_files:
            super().load(cfg_file, return_exceptions=True, ignored_classes=["motors"])
        # VME-style Motors happen later so duplicate motors can be
        # removed
        for cfg_file in config_files:
            super().load(
                cfg_file,
                device_classes={"motors": load_motors},
                ignored_classes=self.device_classes.keys(),
            )
        # Return the final list
        if return_devices:
            return self.devices
        else:
            return self


beamline = HavenInstrument(
    {
        # Ophyd-async devices
        "aerotech_stage": AerotechStage,
        "camera": AravisDetector,
        "energy": EnergyPositioner,
        "high_heat_load_mirror": HighHeatLoadMirror,
        "ion_chamber": IonChamber,
        "kb_mirrors": KBMirrors,
        "motor": Motor,
        "pfcu4": PFCUFilterBank,
        "pss_shutter": PssShutter,
        "sim_detector": SimDetector,
        "table": Table,
        "xspress3": Xspress3Detector,
        "xy_stage": XYStage,
        # Threaded ophyd devices
        "blade_slits": BladeSlits,
        "aperture_slits": ApertureSlits,
        "capillary_heater": CapillaryHeater,
        "power_supply": NHQ203MChannel,
        "synchrotron": ApsMachine,
        "robot": Robot,
        "dxp": make_dxp_device,
        "beamline_manager": BeamlineManager,
        "area_detector": make_area_detector,
        "scaler": MultiChannelScaler,
    },
)

# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2024, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



import asyncio

import pytest
from ophyd_async.testing import set_mock_value

from haven.devices.xia_pfcu import (
    FilterState,
    PFCUFilter,
    PFCUFilterBank,
    PFCUShutter,
    ShutterState,
)


@pytest.fixture()
async def filter_bank(sim_registry):
    bank = PFCUFilterBank(
        prefix="255id:pfcu4:", name="xia_filter_bank", shutters=[(1, 2)]
    )
    await bank.connect(mock=True)
    sim_registry.register(bank)
    yield bank


@pytest.fixture()
async def shutter(filter_bank):
    shutter = filter_bank.shutters[0]
    await asyncio.gather(
        shutter.setpoint.connect(mock=False),
        shutter.readback.connect(mock=False),
        shutter.top_filter.readback.connect(mock=False),
        shutter.bottom_filter.readback.connect(mock=False),
    )
    yield shutter


@pytest.fixture()
async def filter(filter_bank):
    filter = filter_bank.filters[0]
    await filter.readback.connect(mock=False)
    yield filter


def test_shutter_devices(filter_bank):
    """Check that a shutter device is created if
    requested."""
    assert hasattr(filter_bank, "shutters")
    assert isinstance(filter_bank.shutters[0], PFCUShutter)
    assert "fast_shutters" in filter_bank.shutters[0]._ophyd_labels_
    assert (
        filter_bank.shutters[0].top_filter.material.source
        == "mock+ca://255id:pfcu4:filter2_mat"
    )
    assert hasattr(filter_bank, "filters")
    assert isinstance(filter_bank.filters[0], PFCUFilter)
    assert filter_bank.filters[0].material.source == "mock+ca://255id:pfcu4:filter1_mat"
    assert isinstance(filter_bank.filters[3], PFCUFilter)
    # Make sure the shutter blades are not listed as filters
    assert 1 not in filter_bank.filters.keys()
    assert 2 not in filter_bank.filters.keys()


async def test_shutter_signals(shutter):
    # Check initial state
    assert await shutter.top_filter.setpoint.get_value() == False
    assert await shutter.bottom_filter.setpoint.get_value() == False


async def test_shutter_readback(filter_bank, shutter):
    # Set the shutter position
    set_mock_value(filter_bank.readback, "0010")
    # Check that the readback signal gets updated
    assert await shutter.readback.get_value() == ShutterState.OPEN
    # Set the shutter position
    set_mock_value(filter_bank.readback, "0100")
    # Check that the readback signal gets updated
    assert await shutter.readback.get_value() == ShutterState.CLOSED


async def test_shutter_reading(shutter):
    """Ensure the shutter can be read.

    Needed for compatibility with the ``open_shutters_wrapper``.

    """
    assert shutter.readback.name == shutter.name
    reading = await shutter.read()
    assert shutter.name in reading


def test_pfcu_shutter_mask(shutter):
    """A bit-mask used for determining how to set the filter bank."""
    assert shutter.top_mask() == 0b0100
    assert shutter.bottom_mask() == 0b0010


async def test_shutter_open(filter_bank, shutter):
    """If the PFCU filter bank is available, open both blades simultaneously."""
    # Set the other filters on the filter bank
    set_mock_value(filter_bank.readback, "1001")
    # Open the shutter, and check that the filterbank was set
    await shutter.setpoint.set(ShutterState.OPEN)
    assert await filter_bank.setpoint.get_value() == "1011"


async def test_shutter_close(filter_bank, shutter):
    """If the PFCU filter bank is available, open both blades simultaneously."""
    # Set the other filters on the filter bank
    set_mock_value(filter_bank.readback, "1001")
    # Open the shutter, and check that the filterbank was set
    await shutter.setpoint.set(ShutterState.CLOSED)
    assert await filter_bank.setpoint.get_value() == "1101"


async def test_filter_readback(filter):
    set_mock_value(filter._readback, "In")
    assert await filter.readback.get_value() == FilterState.IN


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2024, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------
